"""Exceptions for this package."""


class FieldNameError(Exception):
    """A field name management error occurred."""


class FieldParentError(Exception):
    """A field-to-field-parent management error occurred."""


class InvalidStructObjectError(Exception):
    """Invalid creation of a custom StructObject subclass."""



"""Struct object."""

from collections import OrderedDict
from dataclasses import dataclass
from typing import ClassVar, Sequence, Optional, Mapping, Iterable, Type

from pyspark.sql import types as sql_types
from pyspark.sql.types import DataType, StructField

from ..exceptions import InvalidStructObjectError
from .base import BaseField


@dataclass(frozen=True)
class StructObjectClassMeta:
    """Metadata associated with a struct object; part of the underlying machinery of sparkql."""

    fields: Mapping[str, BaseField]
    spark_struct: sql_types.StructType
    includes: Optional[Sequence["StructObject"]] = None  # ^ TO-DO  https://github.com/mattjw/sparkql/issues/17
    interfaces: Optional[Sequence["StructObject"]] = None  # ^ TO-DO  https://github.com/mattjw/sparkql/issues/16


class StructObject(BaseField):
    """A struct object; shadows StructType in the Spark API."""

    _struct_object_meta: ClassVar[Optional[StructObjectClassMeta]] = None

    #
    # Handle Spark representations for a StructObject object

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return sql_types.StructType

    @property
    def spark_struct_field(self) -> StructField:
        """The Spark StructField for this field."""
        return StructField(
            name=self.field_name, dataType=self._struct_object_meta.spark_struct, nullable=self.is_nullable
        )

    #
    # Hook in to sub-class creation. Ensure fields are pre-processed when a sub-class is declared

    @classmethod
    def __extract_fields(cls) -> Mapping[str, BaseField]:
        fields = OrderedDict((key, value) for key, value in cls.__dict__.items() if isinstance(value, BaseField))
        for field_name, field in fields.items():
            field._contextual_name = field_name  # pylint: disable=protected-access
        return fields

    @staticmethod
    def __build_spark_struct(fields: Iterable[BaseField]) -> sql_types.StructType:
        """Build a Spark struct (StructType) for a list of fields."""
        return sql_types.StructType([field.spark_struct_field for field in fields])

    @classmethod
    def __init_subclass__(cls, **options):  # pylint: disable=unused-argument
        """Hook in to the subclassing of this base class; process fields when sub-classing occurs."""
        super().__init_subclass__()  # pytype: disable=attribute-error

        # Do not re-extract
        if cls._struct_object_meta is not None:
            return

        # Ensure a subclass does not break base class functionality
        for child_prop, child_val in cls.__dict__.items():
            if (child_prop in StructObject.__dict__) and (isinstance(child_val, BaseField)):
                raise InvalidStructObjectError(f"Field should note override inherited class properties: {child_prop}")

        # Extract fields
        fields = cls.__extract_fields()
        cls._struct_object_meta = StructObjectClassMeta(
            fields=fields, spark_struct=StructObject.__build_spark_struct(fields.values())
        )

    #
    # Handle dot chaining for full path ref to nested fields

    def __getattribute__(self, name):
        """
        Customise how field attributes are handled.

        Augment the attribute reference chain to ensure that a field's parent is set.
        """
        prop = super().__getattribute__(name)
        if not name.startswith("_") and isinstance(prop, BaseField):
            return prop.replace_parent(parent=self)
        return prop

    #
    # Other methods

    def __str__(self) -> str:
        """Nicely printable string representation."""
        return (
            f"<{type(self).__name__} \n"
            f"  spark type = {self._spark_type_class.__name__} \n"
            f"  nullable = {self.is_nullable} \n"
            f"  name = {self._resolve_field_name()} <- {[self._name_explicit, self._name_contextual]} \n"
            f"  parent = {self._parent} \n"
            f"  metadata = {self._struct_object_meta}"
            ">"
        )



"""Concrete atomic fields."""

from typing import Type

from pyspark.sql.types import (
    ByteType,
    IntegerType,
    LongType,
    ShortType,
    DecimalType,
    DoubleType,
    FloatType,
    StringType,
    BinaryType,
    BooleanType,
    DateType,
    TimestampType,
    DataType,
)

from .base import AtomicField, IntegralField, FractionalField


#
# DataType -> AtomicType -> NumericType -> IntegralType


class ByteField(IntegralField):
    """Field for Spark's ByteType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return ByteType


class IntegerField(IntegralField):
    """Field for Spark's IntegerType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return IntegerType


class LongField(IntegralField):
    """Field for Spark's LongType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return LongType


class ShortField(IntegralField):
    """Field for Spark's ShortType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return ShortType


#
# DataType -> AtomicType -> NumericType -> FractionalType


class DecimalField(FractionalField):
    """Field for Spark's DecimalType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return DecimalType


class DoubleField(FractionalField):
    """Field for Spark's DoubleType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return DoubleType


class FloatField(FractionalField):
    """Field for Spark's FloatType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return FloatType


#
# DataType -> AtomicType -> non-numeric types


class StringField(AtomicField):
    """Field for Spark's StringType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return StringType


class BinaryField(AtomicField):
    """Field for Spark's BinaryType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return BinaryType


class BooleanField(AtomicField):
    """Field for Spark's BooleanType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return BooleanType


class DateField(AtomicField):
    """Field for Spark's DateType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return DateType


class TimestampField(AtomicField):
    """Field for Spark's TimestampType."""

    @property
    def _spark_type_class(self) -> Type[DataType]:
        return TimestampType



"""Base field and abstract fields."""

from abc import ABC, abstractmethod
from typing import Optional, Type
import copy

from pyspark.sql import types as sql_type
from pyspark.sql.types import StructField, DataType

from ..exceptions import FieldNameError, FieldParentError

# pytype: disable=invalid-annotation


class BaseField(ABC):
    """Root of the field hierarchy; shadows DataType in the Spark API."""

    # Name management logic:
    # - Explicit name (`_name_explicit`): Set via constructor.
    # - Contextual name (`_name_contextual`): Inferred for the field as it is used in a struct object.
    #   This will always get set, although not immediately. The struct object that will contain this field
    #   is responsible for setting the contextual name.
    # The explicit name, if provided, will override the contextual name.

    _nullable: bool = True
    _name_explicit: Optional[str] = None
    _name_contextual: Optional[str] = None
    _parent_struct_object: Optional["StructObject"] = None

    def __init__(self, nullable: bool = True, name: Optional[str] = None):
        """
        Constructor for a base field.

        Args:
            nullable: Is this field nullable.
            name: Field name. If None, field name will be identified via ivar context resolution.
        """
        self._nullable = nullable
        self._name_explicit = name

    #
    # Nullability

    @property
    def is_nullable(self) -> bool:
        """The nullability status of this field."""
        return self._nullable

    #
    # Field path chaining

    @property
    def _parent(self) -> Optional["StructObject"]:
        return self._parent_struct_object

    def replace_parent(self, parent: Optional["StructObject"] = None) -> "BaseField":
        """Return a copy of this Field with the parent attribute set."""
        field = copy.copy(self)
        if self._parent_struct_object is not None:
            raise FieldParentError("Attempted to set parent field that has already been set")
        field._parent_struct_object = parent  # pylint: disable=protected-access
        return field

    #
    # Field name management

    @property
    def _contextual_name(self) -> Optional[str]:
        return self._name_contextual

    @_contextual_name.setter
    def _contextual_name(self, value: str):
        if self._name_contextual is not None:
            raise FieldNameError(
                "Attempted to override a name that has already been set: "
                f"'{value}' replacing '{self._name_contextual}'"
            )
        self._name_contextual = value

    @property
    def field_name(self) -> str:
        """The name for this field."""
        name = self._resolve_field_name()
        if name is None:
            raise FieldNameError(
                "No field name found among: explicit name = {}, inferred name = {}".format(
                    self._name_explicit, self._name_contextual
                )
            )
        return name

    def _resolve_field_name(self) -> Optional[str]:
        """Resolve name for this field, or None if no concrete name set."""
        if self._name_explicit is not None:
            return self._name_explicit
        if self._name_contextual is not None:
            return self._name_contextual
        return None

    #
    # Spark type management

    @property
    @abstractmethod
    def _spark_type_class(self) -> Type[DataType]:
        """The class of the Spark type corresponding to this field."""

    @property
    @abstractmethod
    def spark_struct_field(self) -> StructField:
        """The Spark StructField for this field."""

    #
    # Misc.

    def __str__(self):
        """String formatted object."""
        return (
            f"<{type(self).__name__} \n"
            f"  spark type = {self._spark_type_class.__name__} \n"
            f"  nullable = {self.is_nullable} \n"
            f"  name = {self._resolve_field_name()} <- {[self._name_explicit, self._name_contextual]} \n"
            f"  parent = {self._parent}"
            ">"
        )


class AtomicField(BaseField):
    """
    Atomic field type.

    In the Spark API types hierarchy:

    ```
    DataType
     |- AtomicType
     |- ...
    ```
    """

    @property
    @abstractmethod
    def _spark_type_class(self) -> Type[DataType]:
        """The class of the Spark type corresponding to this field."""

    @property
    def spark_data_type(self) -> sql_type.DataType:
        """Corresponding Spark datatype for this class."""
        return self._spark_type_class()

    @property
    def spark_struct_field(self) -> StructField:
        """The StructField for this object."""
        return StructField(name=self.field_name, dataType=self.spark_data_type, nullable=self.is_nullable)


class NumericField(AtomicField):
    """
    Numeric field type.

    In the Spark API types hierarchy:

    ```
    DataType
     |- AtomicType
         |- NumericType
     |- ...
    ```
    """

    @property
    @abstractmethod
    def _spark_type_class(self) -> Type[DataType]:
        """The class of the Spark type corresponding to this field."""


class IntegralField(NumericField):
    """
    Integral field type.

    In the Spark API types hierarchy:

    ```
    DataType
     |- AtomicType
         |- NumericType
             |- IntegralType
     |- ...
    ```
    """

    @property
    @abstractmethod
    def _spark_type_class(self) -> Type[DataType]:
        """The class of the Spark type corresponding to this field."""


class FractionalField(NumericField):
    """
    Integral field type.

    In the Spark API types hierarchy:

    ```
    DataType
     |- AtomicType
         |- NumericType
             |- FractionalType
     |- ...
    ```
    """

    @property
    @abstractmethod
    def _spark_type_class(self) -> Type[DataType]:
        """The class of the Spark type corresponding to this field."""



"""Obtain path to a field within a possibly nested hierarchy."""

# pylint: disable=protected-access

from typing import Sequence

from pyspark.sql import Column
from pyspark.sql import functions as sql_funcs

from .fields.base import BaseField


def path_seq(field: BaseField) -> Sequence[str]:
    """Items on the path to a field."""
    fields = [field]
    while fields[0]._parent is not None:
        if fields[0].field_name is None:
            raise ValueError("Encountered an unset name while traversing tree")
        fields.insert(0, fields[0]._parent)
    return [f.field_name for f in fields]


def path_str(field: BaseField) -> str:
    """Return dot-delimited path to field `field`."""
    return ".".join(path_seq(field))


def path_col(field: BaseField) -> Column:
    """Return Spark column pointing to field `field`."""
    fields_seq = path_seq(field)
    col: Column = sql_funcs.col(fields_seq[0])  # pylint: disable=no-member
    for field_name in fields_seq[1:]:
        col = col[field_name]
    return col



"""String formatters."""

from pyspark.sql.types import StructType, StructField, ArrayType, DataType


class SparkSchemaPrettifier:
    """Pretty formatting of a Spark schema."""

    PREFIX = "    "

    #
    # Entry point

    @classmethod
    def pretty_schema(cls, struct: StructType) -> str:
        """Prettify `struct`."""
        return cls._pretty_struct_type(struct, depth=0)

    #
    # Helpers

    @classmethod
    def _indent(cls, depth: int) -> str:
        """
        Create indentation prefix for given depth.

        "depth" is defined as the current nestedness within StructTypes. The root (i.e., top level
        StructType) has a depth of zero.
        """
        return cls.PREFIX * (depth + 1)

    @classmethod
    def _boolean_as_str(cls, value: bool) -> str:
        return str(value).lower()

    #
    # Recursive handlers

    @classmethod
    def _pretty_data_type(cls, dtype: DataType, depth: int) -> str:
        """Handle an instance of a data type."""
        if isinstance(dtype, StructType):
            # off load to the struct type handler
            return cls._pretty_struct_type(dtype, depth + 1)
        if isinstance(dtype, ArrayType):
            # off load to the array type handler
            return cls._pretty_array_type(dtype, depth)

        # for all other data types, give the type name
        return dtype.__class__.__name__

    @classmethod
    def _pretty_struct_type(cls, struct: StructType, depth: int) -> str:
        """Handle the case where the data type is a struct type."""
        assert isinstance(struct, StructType)
        if not struct.fields:
            formatted = "StructType(List())"
        else:
            delim = "\n" + cls._indent(depth)
            lines = []
            for field in struct.fields:
                lines.append(f"{delim}{cls._pretty_struct_field(field, depth)}")
            formatted = "StructType(List({}))".format(",".join(lines))
        return formatted

    @classmethod
    def _pretty_array_type(cls, array: ArrayType, depth: int) -> str:
        """Handle the case where the data type is an array type."""
        assert isinstance(array, ArrayType)
        # also print out the `containsNull` ArrayType boolean
        return "ArrayType({},{})".format(
            cls._pretty_data_type(array.elementType, depth), cls._boolean_as_str(array.containsNull)
        )

    @classmethod
    def _pretty_struct_field(cls, field: StructField, depth: int) -> str:
        """Helper for `_pretty_struct_type`, that formats a field."""
        assert isinstance(field, StructField)
        formatted = "StructField({},".format(field.name)
        if isinstance(field.dataType, (StructType, ArrayType)):
            formatted += "\n" + cls._indent(depth + 1)

        formatted += cls._pretty_data_type(field.dataType, depth + 1) + ","

        if isinstance(field.dataType, (StructType, ArrayType)):
            formatted += "\n" + cls._indent(depth + 1)
        formatted += "{})".format(cls._boolean_as_str(field.nullable))

        return formatted


def pretty_schema(struct: StructType) -> str:
    """
    Returns a pretty stringified `struct`.

    This attempts to mimic StructType's default string formatter, but with the
    addition of white space, new lines, and indentation.
    """
    return SparkSchemaPrettifier.pretty_schema(struct)



"""Python Spark SQL DataFrame schema management for sensible humans."""

__version__ = "0.1.0"

from .schema_builder import schema
from .path import path_col, path_seq, path_str
from .formatters import pretty_schema
from .fields import (
    ByteField,
    IntegerField,
    LongField,
    ShortField,
    DecimalField,
    DoubleField,
    FloatField,
    StringField,
    BinaryField,
    BooleanField,
    DateField,
    TimestampField,
    ArrayField,
    StructObject,
)


__all__ = [
    "schema",
    "path_col",
    "path_seq",
    "path_str",
    "pretty_schema",
    "ByteField",
    "IntegerField",
    "LongField",
    "ShortField",
    "DecimalField",
    "DoubleField",
    "FloatField",
    "StringField",
    "BinaryField",
    "BooleanField",
    "DateField",
    "TimestampField",
    "ArrayField",
    "StructObject",
]



"""Array field."""

import copy
from typing import Optional, Generic, TypeVar

from pyspark.sql.types import ArrayType, StructField

from .base import BaseField


ArrayElementType = TypeVar("ArrayElementType", bound=BaseField)


class ArrayField(Generic[ArrayElementType], BaseField):
    """
    Array field; shadows ArrayType in the Spark API.

    A spark schema generated with this array field will behave as follows:
    - Nullability of the `StructField` is given by `ArrayField.nullable`, as per normal.
    - `containsNull` of the `ArrayType` is given by the nullability of the element contained
      within this field. In other words, `ArrayType.containsNull` is given by
      `ArrayField.element.nullable`.

    Attributes:
        etype: Data type info for the element of this array. Should be an instance of a `BaseField`.
    """

    etype: ArrayElementType  # pytype: disable=not-supported-yet

    def __init__(self, element: ArrayElementType, nullable: bool = True, name: Optional[str] = None):
        super().__init__(nullable, name)

        if not isinstance(element, BaseField):
            raise ValueError(f"Array element must be a field. Found type: {type(element)}")

        self.etype = element
        if element._name_explicit is not None:
            raise ValueError("The element field of an array should not have an explicit name")
            # None of the naming mechanics of this array's element type will be used.
            # The name of the element type will not be used for anything

    #
    # Field path chaining

    def replace_parent(
        self, parent: Optional["StructObject"] = None
    ) -> "StructObject":  # pytype: disable=invalid-annotation
        """Return a copy of this array with the parent attribute set."""
        field = copy.copy(self)
        field._parent_struct_object = self.etype.replace_parent(parent=parent)  # pylint: disable=protected-access
        return field

    #
    # Field name management

    @property
    def _contextual_name(self) -> Optional[str]:
        return self._name_contextual

    @_contextual_name.setter
    def _contextual_name(self, value: str):
        self._name_contextual = value
        self.etype._name_contextual = (  # pylint: disable=protected-access
            # set child to same name as parent
            value
        )

    #
    # Spark type management

    @property
    def _spark_type_class(self):
        return ArrayType

    @property
    def spark_struct_field(self) -> StructField:
        """The Spark StructField for this field."""
        # containsNull => is used to indicate if elements in a ArrayType value can have null values
        return StructField(
            name=self.field_name,
            dataType=ArrayType(
                # Note that we do not care about the element's field name here:
                elementType=self.etype.spark_struct_field.dataType,
                containsNull=self.etype.is_nullable,
            ),
            nullable=self.is_nullable,
        )



"""Schema fields."""

from .atomic import (
    ByteField,
    IntegerField,
    LongField,
    ShortField,
    DecimalField,
    DoubleField,
    FloatField,
    StringField,
    BinaryField,
    BooleanField,
    DateField,
    TimestampField,
)
from .array import ArrayField
from .struct import StructObject


__all__ = [
    "ByteField",
    "IntegerField",
    "LongField",
    "ShortField",
    "DecimalField",
    "DoubleField",
    "FloatField",
    "StringField",
    "BinaryField",
    "BooleanField",
    "DateField",
    "TimestampField",
    "ArrayField",
    "StructObject",
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sparkql', 'sparkql.fields']

package_data = \
{'': ['*']}

install_requires = \
['pyspark==2.4.1']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7.0,<0.8.0']}

entry_points = \
{'console_scripts': ['lint = tasks:lint',
                     'reformat = tasks:reformat',
                     'test = tasks:test',
                     'typecheck = tasks:typecheck',
                     'verify-all = tasks:verify_all']}

setup_kwargs = {
    'name': 'sparkql',
    'version': '0.1.1',
    'description': 'sparkql: Apache Spark SQL DataFrame schema management for sensible humans',
    'long_description': None,
    'author': 'Matt J Williams',
    'author_email': 'mattjw@mattjw.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mattjw/sparkql',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6.8,<3.7.0',
}


setup(**setup_kwargs)



"""Schema management operations."""

from typing import Union, Type

from pyspark.sql.types import StructType

from .fields.struct import StructObject


def schema(schema_object: Union[StructObject, Type[StructObject]]) -> StructType:
    """Spark schema for a struct object."""
    return schema_object._struct_object_meta.spark_struct  # pylint: disable=protected-access
